/* This file is part of the KDE project
   Copyright (C) 2006-2007 KovoKs <info@kovoks.nl>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#include <qstringlist.h>
#include <qsocket.h>
#include <qregexp.h>
#include <qtimer.h>

#include <kapplication.h>
#include <kconfig.h>
#include <kdebug.h>
#include <kmessagebox.h>
#include <klocale.h>
#include <kpassdlg.h>
#include <kwallet.h>
using KWallet::Wallet;

#include "db.h"
#include "socketsafe.h"
#include "setup.h"
#include "imap.h"

namespace Mailody {

static inline bool endSimpleCommand(const QString& text)
{
    return text.find(
            QRegExp("(?:^|\n|\r)a02 (?:OK|NO|BAD)[^\n]+(?:\n)\r?$")) > -1;
}

Imap *Imap::m_instance = 0;

Imap::Imap( QWidget* parent,  const char* name)
    : QWidget( parent, name )
    , m_imap(0)
{
    // kdDebug() << k_funcinfo << endl;
    m_instance = this;
    m_db = DB::dbinstance();

    reconnect();

    // This will process the queue with requests that should go to
    // the server
    startTimer(20);

    // This will prevent the autologout after 30 minutes, after each
    // read it will start the timer to send a noop after 25 minutes.
    m_noopTimer = new QTimer( this );
    connect( m_noopTimer, SIGNAL(timeout()),
             SLOT(slotSendNoop()) );

    /*
        // testcases for the regex.

    kdDebug() << "These should be true:" << endl;
    kdDebug() << endSimpleCommand("bla\na02 OK bla\n\r") << endl;   // true
    kdDebug() << endSimpleCommand("a02 OK bla\n")<< endl;          // true
    kdDebug() << endSimpleCommand("bla\na02 OK bla\n\r")<< endl;   // true
    kdDebug() << endSimpleCommand("bla\n\ra02 OK bla\n")<< endl;   // true

    kdDebug() << "These should be false:" << endl;
    kdDebug() << endSimpleCommand("bla a02 OK bla")<< endl;        // false
    kdDebug() << endSimpleCommand("a02 OK bla")<< endl;            // false
    kdDebug() << endSimpleCommand("\n\ra02 OK bla")<< endl;        // false
    kdDebug() << endSimpleCommand("\na02 OK bla")<< endl;          // false
    kdDebug() << endSimpleCommand("bla\na02 OK bla")<< endl;       // false
    kdDebug() << endSimpleCommand("bla\na02 OK bla\n\rbla\n\r")<< endl; // false
    kdDebug() << endSimpleCommand("a02 OK bla\n\rbla\n") << endl;   // false
    kdDebug() << endSimpleCommand("bla\na02 OK bla\r") << endl;   // false

    */
}

Imap::~Imap()
{
    // kdDebug() << k_funcinfo << endl;
    // queued jobs can be removed...
    m_queue.clear();
    m_currentQueueItem=Queue();
}

Imap *Imap::instance()
{
    // kdDebug() << k_funcinfo << endl;
    return m_instance;
}

void Imap::slotLogin()
{
    kdDebug() << k_funcinfo << endl;
    emit status(i18n("Connected"));
    kapp->config()->setGroup("General");
    QString login = kapp->config()->readEntry("userName");
    QString pass;

    Wallet* wallet = Wallet::openWallet(Wallet::NetworkWallet(), this->winId());
    if (wallet && wallet->isOpen() && wallet->hasFolder("mailody"))
    {
        wallet->setFolder( "mailody" );
        wallet->readPassword("account1", pass);
    }
    delete wallet;

    if (pass.isEmpty())
        manualAuth();
    else
    {
        m_queue.prepend(Queue(Queue::Auth, "",
                    "login \"" + quoteIMAP(login)
                    + "\" \"" + quoteIMAP(pass) + "\""));
        m_currentQueueItem=Queue();
        emit statusReady();

        m_readyToSend = true;
    }
}

void Imap::manualAuth()
{
    kdDebug() << k_funcinfo << endl;
    kapp->config()->setGroup("General");
    QString login = kapp->config()->readEntry("userName");

    QCString password;
    int result = KPasswordDialog::getPassword(password,
                i18n("Could not find a valid password, please enter it here"));
    if (!password)
    {
        m_readyToSend = false;
        if (m_imap)
        {
            m_imap->deleteLater();
            m_imap=0;
        }
        emit status(i18n("Offline"));
        return;
    }
    if (result == KPasswordDialog::Accepted)
        m_queue.prepend(Queue(Queue::Auth, "",
                        "login \"" + quoteIMAP(login) + "\" \"" +
                                quoteIMAP(QString(password)) + "\""));
    password.fill(0); // safe enough?
    m_currentQueueItem=Queue();
    emit statusReady();
    m_readyToSend = true;
}

void Imap::slotRead(const QString& received)
{
    // kdDebug() << k_funcinfo << received <<
    //         " for " << m_currentQueueItem.state() << endl;

    m_noopTimer->start( 25*60*1000, true );

    // get the untagged alert...
    if (received.find(
            QRegExp("(?:^|\n|\r)* OK \\[ALERT\\][^\n]+(?:\n)\r?$")) > -1)
    {
        KMessageBox::information(this,
            i18n("Server reported: %1")
                .arg(received.mid(received.find("[ALERT]")+7)));
        m_currentQueueItem=Queue();
        emit statusError(received.mid(received.find("[ALERT]")+7));
        return;
    }

    m_received = received;
    if (m_currentQueueItem.state() == Queue::GetMailBoxList)
        QTimer::singleShot(0, this,
                           SLOT(slotParseGetMailBoxList()));

    else if (m_currentQueueItem.state() == Queue::GetRecent)
        QTimer::singleShot(0, this,
                           SLOT(slotParseGetRecent()));

    else if (m_currentQueueItem.state() == Queue::Move)
        QTimer::singleShot(0, this,
                   SLOT(slotParseMove()));

    else if (m_currentQueueItem.state() == Queue::CheckMail)
        QTimer::singleShot(0, this,
                   SLOT(slotParseCheckMail()));

    else if (m_currentQueueItem.state() == Queue::Expunge)
        QTimer::singleShot(0, this,
                   SLOT(slotParseExpunge()));

    else if (m_currentQueueItem.state() == Queue::SyncMailBox ||
             m_currentQueueItem.state() == Queue::SelectMailBox ||
             m_currentQueueItem.state() == Queue::SaveMessageData)
        QTimer::singleShot(0, this,
                   SLOT(slotParseExists()));

    else if (m_currentQueueItem.state() == Queue::SaveMessage)
        QTimer::singleShot(0,this,
                   SLOT(slotParseSaveMessage()));

    else if (m_currentQueueItem.state() == Queue::Capability)
        QTimer::singleShot(0,this,
                   SLOT(slotParseCapabilities()));

    else if (m_currentQueueItem.state() == Queue::GetHeaderList)
        QTimer::singleShot(0, this,
                   SLOT(slotParseGetHeaderList()));

    else if (m_currentQueueItem.state() == Queue::GetHeaders ||
             m_currentQueueItem.state() == Queue::GetMessage)
        QTimer::singleShot(0, this,
                   SLOT(slotParseGetMessage()));

    else if (m_currentQueueItem.state() == Queue::CreateMailBox)
        QTimer::singleShot(0, this,
                   SLOT(slotParseCreateMailBox()));

    else if (m_currentQueueItem.state() == Queue::DeleteMailBox)
        QTimer::singleShot(0, this,
                   SLOT(slotParseDeleteMailBox()));

    else if (m_currentQueueItem.state() == Queue::RenameMailBox)
        QTimer::singleShot(0, this,
                   SLOT(slotParseRenameMailBox()));

    else if (m_currentQueueItem.state() == Queue::Auth)
    {
        if (m_received.find("a02 NO") != -1)
        {
            m_currentQueueItem=Queue();
            m_readyToSend = false;

            emit statusError(i18n("Offline"));
            int i = KMessageBox::questionYesNoCancel(this,
                    i18n("The server refused the supplied username and password, "
                         "do you want to go to the settings, re-enter it for one "
                         "time or do nothing?"),
                    i18n("Could not log in"),
                    i18n("Settings"), i18n("Single Input"));
            if (i == KMessageBox::Yes)
            {
                Setup setup(this,"Setup");
                if (setup.exec() != QDialog::Accepted)
                    return;
                slotLogin();
            }
            else if (i == KMessageBox::No)
                manualAuth();
            else
            {
                if (m_imap)
                {
                    m_imap->deleteLater();
                    m_imap=0;
                }
                return;
            }
        }

        emit connected();
        m_currentQueueItem = Queue();
        QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
    }

    else if (m_currentQueueItem.state() == Queue::Connecting)
    {
        // eat. We need to wait for a connected signal.
    }
    else
    {
        m_currentQueueItem = Queue();
        QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
    }
}


// -------------------- Parsers -------------------------------------//

void Imap::slotParseGetMailBoxList()
{
    // kdDebug() << k_funcinfo << endl;

    static QString all_data;
    all_data.append(m_received);

    if ( !endSimpleCommand(all_data) )
        return;

    // Microsoft Server returns quotes sometimes:
    // * LIST (\Marked \HasNoChildren) "/" bbullshit
    // * LIST (\Marked \HasNoChildren) "/" "Trash/Deleted Items"

    QRegExp rx1("LIST \\(.*\\) \".\" (.*)\n");
    rx1.setMinimal(true);

    QMap<QString, bool> newList;
    int start = 0;
    while (start != -1)
    {
        start = rx1.search(all_data, start);

        if (start == -1)
            break;

        QString y = rx1.cap(1).stripWhiteSpace();

        // Remove them if they are there.... Think MS.
        if (y.startsWith("\"") && y.endsWith("\""))
            y=y.mid(1, y.length()-2);

        // kdDebug() << "Mailbox found.. " << y << endl;
        newList[y] = true;

        ++start;
    }

    m_db->beginTransaction();

    // Get existing mailboxes.
    QStringList list = m_db->getMailBoxList();
    QStringList::Iterator it = list.begin();
    while (it != list.end())
    {
        QString oldItem = (*it);
        ++it;

        // check if it exists in the new list.
        if (newList[oldItem])
            newList[oldItem] = false;     // prevent later addition
        else
        {
            // so its not in the new list, we need to remove it now...
            kdDebug() << "Removing: " << oldItem  << endl;
            m_db->deleteMessagesAndMailBoxes(oldItem);
        }
        ++it;   // skip count;
    }

    // For the newList add them if still true.
    QMap<QString,bool>::Iterator ita;
    for (ita = newList.begin(); ita != newList.end(); ++ita)
    {
        if (ita.data())
        {
            // kdDebug() << "Adding: " << ita.key() << endl;
            m_db->insertMailBox(ita.key());
        }
    }

    m_db->commitTransaction();
    all_data = QString::null;
    emit mailBoxListUpdated();
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseGetRecent()
{
    // kdDebug() << k_funcinfo << endl;

    QRegExp rx("SEARCH (.*)a02");
    if (rx.search(m_received.stripWhiteSpace()) != -1)
    {
        QString r = rx.cap(1).stripWhiteSpace();
        QStringList t =
            QStringList::split(" ",r);

        emit unseenCount(m_currentQueueItem.mailbox(), t.count());

        QStringList all;
        m_db->getCurrentMessageIDs(m_currentQueueItem.mailbox(), all);
        QMap<int, bool> allnum;
        QStringList::iterator it = all.begin();
        while(it != all.end())
        {
            allnum[(*it).toInt()] = true;
            ++it;
        }

        QStringList fetchList;
        QStringList::iterator it2 = t.begin();
        while(it2 != t.end())
        {
            if (!allnum[(*it2).toInt()])
                fetchList.append(*it2);
            ++it2;
        }

        getHeaders(fetchList, m_currentQueueItem.mailbox());
        // kdDebug() << all << " - " << fetchList.join(",") << endl;
    }
    else
        emit unseenCount(m_currentQueueItem.mailbox(), 0);

    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseMove()
{
    // kdDebug() << k_funcinfo << endl;

    QRegExp rx("a02 OK");
    if (rx.search(m_received.stripWhiteSpace()) == -1)
    {
        kdDebug() << "Store failed, removing the delete command" << endl;
        m_queue.pop_front();
    }
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseCheckMail()
{
    // kdDebug() << k_funcinfo << endl;

    // kdDebug() << "checkmailresult: " << m_currentQueueItem.mailbox()
    //        << " : " << m_received << endl;
    QRegExp rx1("STATUS (.+) \\(MESSAGES (\\d+) UNSEEN (\\d+)\\)");
    rx1.setMinimal(true);

    QString mb;
    QString amount;
    int totalShouldBe=0;
    int unseenShouldBe=0;

    if (rx1.search(m_received.stripWhiteSpace()) != -1)
    {
        mb = rx1.cap(1);

        //regex in qt is weird...
        if (mb.startsWith("\"") && mb.endsWith("\""))
            mb = mb.mid(1,mb.length()-2);

        amount = rx1.cap(2);

        m_db->setTotalMessagesMailbox(m_currentQueueItem.mailbox(),
                                      amount.toInt());
        emit messageCount(m_currentQueueItem.mailbox(), amount);
        totalShouldBe = amount.toInt();

        unseenShouldBe = rx1.cap(3).toInt();
        emit unseenCount(m_currentQueueItem.mailbox(), unseenShouldBe);
    }
    else
    {
        // we appereantly m_received garbage, should not happen, but
        // there are more things in life, requeue and get on with it.
        // possible loop of course...
        // kdDebug() << "unexpected response received" << m_received << endl;
        m_queue.append(m_currentQueueItem);
        m_currentQueueItem = Queue();
        QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
        emit statusError(i18n("Something is wrong..."));
        return;
    }
    integrity(m_currentQueueItem.mailbox(), totalShouldBe);
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::integrity(const QString& mb, int totalShouldBe)
{
    // kdDebug() << k_funcinfo << endl;
    int mailsReal = m_db->getTotalMessagesMessages(mb);

    // kdDebug() << "integrity: " << mb << " should be: "
    //    << totalShouldBe << " current: " << mailsReal << endl;

    if (totalShouldBe > mailsReal)
    {
        // kdDebug() << "Fetching " << mailsReal+1 << " - "
        //     << totalShouldBe << endl;
        getHeaderList(mailsReal+1, totalShouldBe, mb);
    }
    else if (totalShouldBe != mailsReal)
    {
        // kdDebug() << "folder dirty: on server: " << totalShouldBe <<
        //        " - in database: " << mailsReal << endl;
        m_db->deleteMessages(mb);
        getHeaderList(1, totalShouldBe, mb);

        // clear the view!
        emit allUidsKnown( mb );
    }
//    else
//        kdDebug() << "Fine" << endl;
}

void Imap::slotParseCapabilities()
{
    // kdDebug() << k_funcinfo << " : " << m_received << endl;
    m_capabilities = QStringList::split(" ", m_received.lower());
    capable();  // that was the reason we ended up here.

    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseExpunge()
{
    if (m_currentQueueItem.state() == Queue::Expunge)
    {
        // Delete them from the database..
        m_db->expunge( m_currentQueueItem.mailbox() );

        // Update the view.
        emit allUidsKnown( m_currentQueueItem.mailbox() );
    }
    QTimer::singleShot(0, this,
                   SLOT(slotParseExists()));
}

void Imap::slotParseSaveMessage()
{
    // kdDebug() << k_funcinfo << " : " << m_received << endl;
    m_currentQueueItem = m_queue.first();
    if (m_imap)
            m_imap->write( m_currentQueueItem.command() );
    m_queue.pop_front();
}

void Imap::slotParseCreateMailBox()
{
    // kdDebug() << k_funcinfo << " : " << m_received << endl;

    QRegExp rx("a02 OK");
    if (rx.search(m_received.stripWhiteSpace()) == -1)
    {
        kdDebug() << "Failed to create folder" << endl;
    }
    else
    {
        QString box = m_currentQueueItem.mailbox();
        // kdDebug() << "Adding: " << box << endl;
        m_db->insertMailBox(box);
        emit mailBoxListUpdated();
    }
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseDeleteMailBox()
{
    // kdDebug() << k_funcinfo << " : " << m_received << endl;

    QRegExp rx("a02 OK");
    if (rx.search(m_received.stripWhiteSpace()) == -1)
    {
        kdDebug() << "Failed to delete folder" << endl;
    }
    else
    {
        QString box = m_currentQueueItem.mailbox();
        // kdDebug() << "Deleting: " << box << endl;
        m_db->deleteMessagesAndMailBoxes(box);
        emit mailBoxListUpdated();
    }
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseRenameMailBox()
{
    // kdDebug() << k_funcinfo << " : " << m_received << endl;

    QRegExp rx("a02 OK");
    if (rx.search(m_received.stripWhiteSpace()) == -1)
    {
        kdDebug() << "Failed to rename folder" << endl;
    }
    else
    {
        QString boxes = m_currentQueueItem.mailbox();
        QStringList i = QStringList::split("~",boxes);

        // kdDebug() << "renaming: " << i[0] << "->" << i[1] << endl;
        m_db->renameMailBox(i[0],i[1]);
        emit mailBoxListUpdated();
    }
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseExists()
{
    // kdDebug() << k_funcinfo << " for " << m_received << endl;

    m_selectedMailbox=m_currentQueueItem.mailbox();

    // Error handling
    if (m_received.find("a02 NO") != -1)
    {
        // if we get here because the queue was delayed, then we
        // need to remove it from the queue to prevent a loop.
        Queue nextQueueItem = m_queue.first();
        if (m_currentQueueItem.mailbox() == nextQueueItem.mailbox())
        {
            kdDebug() << "Removing this and next command" << endl;
            m_queue.pop_front();
        }

        emit statusError(i18n("Failed to select the mailbox"));
        m_currentQueueItem = Queue();
        QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
        return;
    }

    QRegExp rx1(" (\\d+) EXISTS");
    if (rx1.search(m_received.stripWhiteSpace()) != -1)
    {

        QString amount = rx1.cap(1);
        m_db->setTotalMessagesMailbox(m_currentQueueItem.mailbox(),
                                      amount.toInt());
        emit messageCount(m_currentQueueItem.mailbox(), amount);

        if (m_currentQueueItem.state() == Queue::SyncMailBox)
            getHeaderList(1, amount.toInt(), m_selectedMailbox );
        else
            integrity(m_selectedMailbox, amount.toInt());
    }

    // Sync will not emit the unseen count, so do a checkmail after that.
    if (m_currentQueueItem.state() == Queue::SyncMailBox)
        checkMail(m_selectedMailbox);

    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseGetHeaderList()
{
    // kdDebug() << k_funcinfo << endl;

    static QString all_data;
    all_data.append(m_received);

    if ( !endSimpleCommand(all_data) )
        return;

    // * 1 FETCH (UID 1 FLAGS (\Seen))                  // CYRUS
    // * 3 FETCH (FLAGS (\Deleted \Seen) UID 26135)     // Courier
    QRegExp rx1("UID [\\(]?(\\d*)[\\)]?");
    QRegExp rx2("FLAGS \\((.*)\\)");
    rx2.setMinimal(true);

    QStringList fetchlist, allList, updateList;

    // Get all uids, so we dont have to query
    QStringList all;
    m_db->getCurrentMessageIDs(m_currentQueueItem.mailbox(), all);
    QMap<int, bool> allnum;
    QStringList::iterator it = all.begin();
    while(it != all.end())
    {
        allnum[(*it).toInt()] = true;
        ++it;
    }

    all_data = all_data.stripWhiteSpace();

    m_db->beginTransaction();
    int start = 0, start2 = 0;
    while (start != -1)
    {
        start = rx1.search(all_data, start);
        if (start == -1)
            break;

        int uid = rx1.cap(1).toInt();
        if (! uid > 0)
        {
            ++start;
            continue;
        }

        if (!allnum[uid])
            fetchlist.append(rx1.cap(1));

        start2 = rx2.search(all_data, start2);
        if (start2 != -1)
            m_db->setFlags(uid, m_currentQueueItem.mailbox(), rx2.cap(1));

        ++start2;
        ++start;
    }
    m_db->commitTransaction();

    all_data = QString::null;
    getHeaders(fetchlist, m_currentQueueItem.mailbox());
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotParseGetMessage()
{
    // kdDebug() << k_funcinfo << endl;

    // Courier:
    //|* 1 FETCH (UID 1 RFC822.SIZE 8609 BODY[HEADER.FIELDS ("FROM" "SUBJECT" "DATE")] {148}|
    // Cyrus:
    //|* 1 FETCH (UID 26133 RFC822.SIZE 8609 BODY[HEADER.FIELDS (FROM SUBJECT DATE)] {139}|
    // Dovecot:
    //|* 1 FETCH (RFC822.SIZE 8609 UID 2 BODY[HEADER.FIELDS (FROM TO CC SUBJECT DATE IN-REPLY-TO MESSAGE-ID)] {337}

    static QString all_data;
    all_data.append(m_received);

    if ( !endSimpleCommand(all_data) )
        return;

    QRegExp rx0;
    QRegExp rx1;
    QRegExp rx2;
    if (m_currentQueueItem.state() == Queue::GetHeaders)
    {
        rx0.setPattern("UID (\\d*)");
        rx1.setPattern("RFC822.SIZE (\\d*)");
        rx2.setPattern( QRegExp::escape(" BODY[HEADER.FIELDS ")
                + "\\(\"?FROM\"? \"?TO\"? \"?CC\"? "
                + "\"?SUBJECT\"? \"?DATE\"? \"?IN-REPLY-TO\"? "
                + "\"?MESSAGE-ID\"?\\)\\] "
                + QRegExp::escape("{") + "(\\d*)" + QRegExp::escape("}"));
    }
    else
    {
        rx0.setPattern("UID (\\d*)");
        rx2.setPattern(QRegExp::escape("BODY[] ")
                + QRegExp::escape("{") + "(\\d*)" + QRegExp::escape("}"));
    }

    m_db->beginTransaction();
    QStringList headersToSend;

    while (rx2.search(all_data.stripWhiteSpace()) != -1)
    {
        // First capture the body part:
        QString size = rx2.cap(1);
        int startBody = rx2.pos(1)+size.length()+2;
        QString text = all_data.mid(startBody, size.toInt());

        // Second, remove the complete regex part from the string.
        all_data = all_data.left(rx2.pos(0))+all_data.mid(startBody+size.toInt());

        // Now figure out the uid.
        int uid=0;
        if (rx0.search(all_data.stripWhiteSpace()) != -1)
            uid = rx0.cap(1).toInt();

        if (m_currentQueueItem.state() == Queue::GetHeaders)
        {
            QString size;
            if (m_currentQueueItem.state() == Queue::GetHeaders &&
                rx1.search(all_data.stripWhiteSpace()) != -1)
                size = "Size: " + rx1.cap(1);

            text.append("\r\n");
            text.append(size);

            headersToSend.append(rx0.cap(1));
            headersToSend.append(m_currentQueueItem.mailbox());
            headersToSend.append(text);
            headersToSend.append(m_db->getFlags(uid,
                                 m_currentQueueItem.mailbox()));

            if (text.isEmpty())
                text = "Message corrupt";

            m_db->storeHeaders(uid, m_currentQueueItem.mailbox(),text);
        }
        else
        {
            // kdDebug() << "storing" << endl;
            m_db->storeBody(uid,m_currentQueueItem.mailbox(),text);
            emit bodyAvailable(m_currentQueueItem.mailbox(), uid);
        }

        // Find the next occorunce of FETCH and strip untill there.
        all_data = all_data.mid(all_data.find("FETCH", rx0.pos(1),false));
    }
    m_db->commitTransaction();

    all_data = QString::null;
    if (m_currentQueueItem.state() == Queue::GetHeaders)
        emit addHeaders(m_currentQueueItem.mailbox(), headersToSend );
    // kdDebug() << "done" << endl;
    emit statusReady();
    m_currentQueueItem = Queue();
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

// ------------ Mailbox operations ----------------------- //

void Imap::getMailBoxList(bool complete)
{
    // kdDebug() << k_funcinfo << endl;
    if (complete)
        m_db->deleteMessagesAndMailBoxes();

    m_queue.append(Queue(Queue::GetMailBoxList,"", "LIST \"\" \"*\""));
}

void Imap::syncMailBox(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_db->deleteMessages( box );
    emit messageCount( box, 0);
    emit unseenCount( box, 0);
    emit allUidsKnown( box ); // clear the view
    m_queue.prepend(Queue(Queue::SyncMailBox, box, "SELECT \""+box+"\""));
}

void Imap::selectMailBox(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_queue.prepend(Queue(Queue::SelectMailBox, box, "SELECT \""+box+"\""));
}

void Imap::checkMail(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_queue.append(Queue(Queue::CheckMail, box,
                   "STATUS \""+box+"\" (MESSAGES UNSEEN)"));
}

void Imap::slotSendNoop()
{
    kdDebug() << k_funcinfo << endl;
    m_queue.append(Queue(Queue::NoResponse, "","NOOP"));
}

void Imap::expungeMailBox(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_queue.append(Queue(Queue::Expunge, box, "EXPUNGE"));
}

void Imap::createMailBox(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_queue.append(Queue(Queue::CreateMailBox, box, "CREATE \"" + box + "\""));
}

void Imap::deleteMailBox(const QString& box)
{
    // kdDebug() << k_funcinfo << box << endl;
    m_queue.append(Queue(Queue::DeleteMailBox, box, "DELETE \"" + box + "\""));
}

void Imap::renameMailBox(const QString& oldbox, const QString& newbox)
{
    // kdDebug() << k_funcinfo << oldbox << "->" << newbox << endl;
    m_queue.append(Queue(Queue::RenameMailBox, oldbox + "~" + newbox,
                   "RENAME \"" + oldbox + "\" \"" + newbox + "\""));
}

//----------------   Messages -----------------------------//

void Imap::moveMessage(const int uid, const QString& origbox,
                       const QString& destbox)
{
    // kdDebug() << k_funcinfo << origbox
    //    << " - " << uid << " -> " << destbox << endl;
    m_queue.append(Queue(Queue::Move, origbox,
                   "UID COPY " + QString::number(uid) + " \"" + destbox+ "\""));
    addFlag(uid, origbox, "\\Deleted");
}

void Imap::addFlag(const int uid, const QString& box, const QString& flag)
{
    // kdDebug() << k_funcinfo << box << " - " << uid << " - " << flag << endl;
    m_queue.append(Queue(Queue::NoResponse, box,
                   "UID STORE " + QString::number(uid)
                           + " +FLAGS (" + flag + ')'));
    m_db->addFlag(uid, box, flag);
}

void Imap::addFlag(const QString& box, const QString& flag)
{
    // kdDebug() << k_funcinfo << box << " - " << uid << " - " << flag << endl;
    int min = 0;
    int max = 0;
    m_db->getMinMax(box,min,max);
    m_queue.append(Queue(Queue::NoResponse, box,
                   "UID STORE "
                           + QString::number(min) + ":" + QString::number(max)
                           + " +FLAGS (" + flag + ')'));
    m_db->addFlag(box, flag);
}

void Imap::removeFlag(const int uid, const QString& box, const QString& flag)
{
    // kdDebug() << k_funcinfo << box << " - " << uid << " - " << flag << endl;
    m_queue.append(Queue(Queue::NoResponse, box,
                   "UID STORE " + QString::number(uid)
                           + " -FLAGS (" + flag + ')'));
    m_db->removeFlag(uid, box, flag);
}

void Imap::getBody(int uid, const QString& mb)
{
    // kdDebug() << k_funcinfo << endl;
    if (!m_db->hasBody(uid, mb))
        m_queue.prepend( Queue(Queue::GetMessage, mb,
                         "UID FETCH " + QString::number(uid)
                                 + ':' + QString::number(uid) + " BODY[]"));
    else
        emit bodyAvailable(mb, uid);
}

void Imap::saveMessage(const QString& mb, const QString& message)
{
    // kdDebug() << k_funcinfo << message << endl;
    m_queue.append( Queue(Queue::SaveMessage, mb,
                    "APPEND \"" + mb + "\" (\\Seen) {"
                            + QString::number(message.length())
                            + "}"));
    m_queue.append( Queue(Queue::SaveMessageData, mb, message));
}

//------------------------- headers -------------------------------//

void Imap::getHeaderList(int start, int end, const QString& mb)
{
    // kdDebug() << k_funcinfo << start << "-" << end
    //        << "for" << mb << endl;
    if (! end>=1 || end < start || !start>=1)
        return;

    // Put this in queue with priority above the checkmail things...
    m_queue.prepend(Queue(Queue::GetHeaderList, mb,
                   "FETCH " + QString::number(start) + ':'
                           + QString::number(end) + " (UID FLAGS)"));
}

void Imap::getHeaders(const QStringList& uids, const QString& mb)
{
    // kdDebug() << k_funcinfo << << uids << mb << endl;
    if (uids.count() == 0 || mb.isEmpty())
        return;

    // Join first, then split it in blocks of 250.
    QString complete = uids.join(",");

    while (!complete.isEmpty())
    {
        QString part = complete.section(',',0,250);
        complete = complete.section(',',251);

        // Priority; above the checkmail calls in the queue.
        m_queue.prepend(Queue(Queue::GetHeaders, mb,
                   "UID FETCH " + part
           + " (RFC822.SIZE BODY.PEEK[HEADER.FIELDS (FROM TO CC SUBJECT "
                           "DATE IN-REPLY-TO MESSAGE-ID)])"));
    }
}

// ----------------------------- Others --------------------------//

void Imap::getCapabilities()
{
    kdDebug() << k_funcinfo << endl;
    m_queue.append( Queue(Queue::Capability, QString::null, "CAPABILITY"));
}

void Imap::capable(const QString& something)
{
    kdDebug() << k_funcinfo << something << endl;

    static QString s_savedSomething;

    if (!m_capabilities.count())
    {
        getCapabilities();
        s_savedSomething = something.lower();
        return;
    }

    if (!something.isEmpty())
        s_savedSomething = something.lower();

    if (s_savedSomething.isEmpty())
        return;

    bool result = m_capabilities.findIndex( s_savedSomething ) > -1;
    kdDebug() << "capable of " << s_savedSomething << ": " << result << endl;

    emit capable(s_savedSomething, result);
    s_savedSomething = QString::null;
}

// ----------------------------- IDLE ---------------------------//

void Imap::idleStart(const QString& mb)
{
    kdDebug() << k_funcinfo << endl;
    m_queue.append( Queue(Queue::IdleStart, mb, "IDLE"));
}

void Imap::idleStop()
{
    kdDebug() << k_funcinfo << endl;
    m_queue.prepend( Queue(Queue::IdleStop, QString::null, "DONE"));
}

// ---------------------------- Management -----------------------//

void Imap::write(const QString& text)
{
    // kdDebug() << k_funcinfo << text << endl;
    if (m_imap)
        m_imap->write("a02 "+text);
}

void Imap::slotProcessQueue()
{
    /*
        kdDebug() << m_queue.count() << " - " << m_readyToSend
             << " - " << m_imap->available()
             << " - " << m_currentQueueItem.state()  << endl;
    */

    if (!m_readyToSend || !m_imap || !m_queue.count() ||
        m_currentQueueItem.state() != Queue::None ||
        !m_imap->available() )
        return;

    m_currentQueueItem = m_queue.first();

    if (m_selectedMailbox == m_currentQueueItem.mailbox() ||
        m_currentQueueItem.mailbox().isEmpty() ||
        m_currentQueueItem.state() == Queue::SyncMailBox ||
        m_currentQueueItem.state() == Queue::SelectMailBox ||
        m_currentQueueItem.state() == Queue::CreateMailBox ||
        m_currentQueueItem.state() == Queue::DeleteMailBox ||
        m_currentQueueItem.state() == Queue::RenameMailBox ||
        m_currentQueueItem.state() == Queue::CheckMail)
    {
        if (m_currentQueueItem.state() == Queue::CheckMail)
        {
            if (m_currentQueueItem.mailbox() == m_selectedMailbox)
                m_currentQueueItem =
                        Queue(Queue::GetRecent, m_selectedMailbox,
                              "UID SEARCH UNSEEN");
            emit status(i18n("Checking mail for %1")
                    .arg(m_currentQueueItem.mailbox()));
        }
        else if (m_currentQueueItem.state() == Queue::GetHeaderList)
            emit status(i18n("Retrieving status list for %1")
                    .arg(m_currentQueueItem.mailbox()));
        else if (m_currentQueueItem.state() == Queue::GetHeaders)
            emit status(i18n("Retrieving headers for %1")
                    .arg(m_currentQueueItem.mailbox()));
        else if (m_currentQueueItem.state() == Queue::GetMailBoxList)
            emit status(i18n("Retrieving mailbox list"));
        else if (m_currentQueueItem.state() == Queue::GetMessage)
            emit status(i18n("Retrieving message"));
        else if (m_currentQueueItem.state() == Queue::Move)
            emit status(i18n("Moving message"));

        write( m_currentQueueItem.command() );
        m_queue.pop_front();
    }
    else
    {
        // We need to switch to that mb first
        /*kdDebug() << "Delayed, first selecting "<< m_currentQueueItem.mailbox()
                << " instead of " << m_selectedMailbox << " -> for: "
        << m_currentQueueItem.state() << endl; */
        selectMailBox(m_currentQueueItem.mailbox());
        m_currentQueueItem=Queue();
    }
}

void Imap::timerEvent( QTimerEvent * )
{
    QTimer::singleShot(0, this, SLOT(slotProcessQueue()));
}

void Imap::slotError(const QString& msg)
{
    kdDebug() << k_funcinfo << msg << endl;
    m_imap->aboutToClose();
    emit status(i18n("Offline"));

    KMessageBox::information(this, msg);
}

void Imap::slotShowDisConnectedError()
{
    // kdDebug() << k_funcinfo << endl;
    kdDebug() << "slotDisconnected" << endl;

    /* Erase the connection data since at this point we are not
       connected anywhere. */
    m_server = QString::null;
    m_port = -1;
    m_safe = -1;

    emit status(i18n("Offline"));

    int i = KMessageBox::questionYesNo(0,
                            i18n("Connection has been closed, reconnect?"));
    if (i == KMessageBox::Yes)
        reconnect();
}

void Imap::slotTLS()
{
    kdDebug() << k_funcinfo << endl;
    m_imap->write("a02 starttls");
}

bool Imap::reconnect()
{
    // kdDebug() << k_funcinfo << endl;
    kapp->config()->setGroup("General");
    const QString imap_server = kapp->config()->readEntry("imapServer");
    int safe = kapp->config()->readNumEntry("safeImap",2);

    QString server = imap_server.section(":",0,0);
    int port = imap_server.section(":",1,1).toInt();
    bool skip_reconnect = false;

    /* Determine if the user needs to reconnect. */
    if (m_server)
        skip_reconnect = (safe == m_safe && server == m_server &&
                          port == m_port);

    kdDebug() << "skip_reconnect = " << skip_reconnect << endl;

    if (skip_reconnect)
        return false;
    else
    {
        /* preserve data to disallow skip reconnecting next time. */
        m_server = server;
        m_safe = safe;
        m_port = port;
    }

    m_selectedMailbox = QString::null;
    m_currentQueueItem = Queue();
    m_readyToSend = false;
    m_queue.clear();

    if (m_imap)
    {
        kdDebug() << "Delete reconnect" << endl;
        m_imap->aboutToClose();
        m_imap->deleteLater();
        m_imap = 0;
    }

    switch(safe)
    {
        case 0:
            if (port == 0)
                port = 143;
            m_imap  = new SocketSafe(this,"IMAP",server, port, SocketSafe::NONE);
            break;
        case 1:
            if (port == 0)
                port = 993;
            m_imap  = new SocketSafe(this,"IMAP",server, port, SocketSafe::SSL);
            break;
        case 2:
            if (port == 0)
                port = 143;
            m_imap = new SocketSafe(this,"IMAP",server, port, SocketSafe::TLS);
            break;
    }

    connect(m_imap, SIGNAL(data(const QString&)),
            SLOT(slotRead(const QString&)));
    connect(m_imap, SIGNAL(connected()), SLOT(slotLogin()));
    connect(m_imap, SIGNAL(error(const QString&)),
            SLOT(slotError(const QString&)));
    connect(m_imap, SIGNAL(disconnected()),
            SLOT(slotShowDisConnectedError()));
    connect(m_imap, SIGNAL(readyfortls()),
            SLOT(slotTLS()));

    emit status(i18n("Connecting"));
    m_currentQueueItem= Queue(Queue::Connecting,"","");
    m_imap->reconnect();
    return true;
}

/* Copy from KDE/3.5/kdepim/kioslaves/imap4/rfcdecoder.cc */
/* Copyright (C) 2000 s.carstens@gmx.de */
/* replace " with \" and \ with \\ " and \ characters */
QString Imap::quoteIMAP(const QString &src)
{
    uint len = src.length();
    QString result;
    result.reserve(2 * len);
    for (unsigned int i = 0; i < len; i++)
    {
        if (src[i] == '"' || src[i] == '\\')
            result += '\\';
        result += src[i];
    }
    //result.squeeze(); - unnecessary and slow
    return result;
}

}

#include "imap.moc"
