/* This file is part of the KDE project

   Copyright (C) 2006-2007 KovoKs <info@kovoks.nl>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public
   License as published by the Free Software Foundation; either
   version 2 of the License, or (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
   General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/


#ifndef IMAP_H
#define IMAP_H

#include "qwidget.h"
class QSocket;

namespace Mailody {

class DB;

class Queue
{
    public:
        enum Commands{
            None=0,
            NoResponse,
            Connecting,
            Auth,
            GetMailBoxList,
            CheckMail,
            SelectMailBox,
            SyncMailBox,
            GetHeaderList,
            GetHeaders,
            GetMessage,
            GetRecent,
            Move,
            Expunge,
            SaveMessage,
            SaveMessageData,
            CreateMailBox,
            DeleteMailBox,
            RenameMailBox,
            Capability,
            IdleStart,
            IdleStop
        };

        Queue(): st(0) {}
        explicit Queue( Commands state, const QString& mailbox,
               const QString& command )
            : st(state), mb(mailbox), com(command)
        { }
        int state() const {return st; }
        QString mailbox() const {return mb; }
        QString command() const {return com; }

    private:
        int st;
        QString mb, com;
};


class SocketSafe;

/**
 * @class Imap
 * Responsible for communicating with the server and storing stuff in the
 * database.
 * @author Tom Albers <tomalbers@kde.nl>
 */
class Imap : public QWidget
{
  Q_OBJECT

  public:
    /**
     * Contructor
     */
    explicit Imap( QWidget* parent,  const char* name );

    /**
     * Destructor
     */
    ~Imap();

    /**
     * Will fetch the mailbox list and store it in the db.
     * @param complete if true, it will delete every mailbox
     * and all messages before it starts.
     */
    void getMailBoxList(bool complete);

    /**
     * Use this to sync a mailbox, all headers will be fetched.
     * @param box the mailbox
     */
    void syncMailBox(const QString& box);

    /**
     * This will select the mailbox and tries to fetch only whats needed.
     * @param box the mailbox
     */
    void selectMailBox(const QString& box);

    /**
     * This will not select the mailbox, but will update message count
     * and fetch new mail
     * @param box the mailbox
     */
    void checkMail(const QString& box);

    /**
     * add a flag, goes to the server and to the database for a uid in a box
     * @param uid uid of the message
     * @param box the mailbox
     * @param flag flag to add
     */
    void addFlag(const int uid, const QString& box, const QString& flag);

    /**
     * add a flag, goes to the server and to the database. for a whole box
     * @param box the mailbox
     * @param flag flag to add
     */
    void addFlag(const QString& box, const QString& flag);

    /**
     * remove a flag, goes to the server and to the database.
     * @param uid uid of the message
     * @param box the mailbox
     * @param flag flag to add
     */
    void removeFlag(const int uid, const QString& box, const QString& flag);

    /**
     * removes deleted messages
     * @param box the mailbox
     */
    void expungeMailBox(const QString& box);

    /**
     * create a folder
     */
    void createMailBox(const QString& box);

    /**
     * delete a folder
     */
    void deleteMailBox(const QString& box);

    /**
     * rename folder @p oldfolder to @p newfolder
     */
    void renameMailBox(const QString& oldfolder, const QString& newfolder);

    /**
     * This will move the message with @p uid from @p origbox to the
     * @p destbox mailbox.
     */
    void moveMessage(const int uid, const QString& origbox,
                     const QString& destbox);

    /**
     * Get the headers with the @p uids from @p mb mailbox
     */
    void getHeaders(const QStringList& uids, const QString& mb);

    /**
     * Get the body of @p uid from the @p mb mailbox
     */
    void getBody(int uid, const QString& mb);

    /**
     * Store message with the content @p message in the @p mb mailbox
     */
    void saveMessage(const QString& mb, const QString& message);

    /**
     * Reconnect to the imap server. You can call this safely with
     * changing from socket to secure socket and vice versa
     * @returns true when there is a new connection, false when there
     * is no need for a new connection.
     */
    bool reconnect();

    /**
     * Get the capabilities of the server, the result will be told via the
     * capable signal.
     */
    void capable(const QString& something=QString::null);

    /**
     * Start monitoring the @p mb mailbox
     */
    void idleStart(const QString& mb);

    /**
     * Stop monitoring
     */
    void idleStop();

    /**
     * Returns an instance to this class. Make sure there is one.
     */
    static Imap *instance();

 private:

    SocketSafe*         m_imap;
    Queue               m_currentQueueItem;
    QValueList<Queue>   m_queue;
    DB*                 m_db;
    QString             m_selectedMailbox;
    bool                m_readyToSend;
    QTimer*             m_noopTimer;
    static Imap*        m_instance;
    QString             m_received;
    void manualAuth();
    QStringList         m_capabilities;
    /* Saves the current parameters of the connection. */
    QString             m_server;
    int                 m_port;
    int                 m_safe;

    void write(const QString& text);
    void getHeaderList(int start, int end, const QString& mb);
    void integrity(const QString& mb, int);
    void timerEvent( QTimerEvent * );
    void getCapabilities();
    QString quoteIMAP(const QString &src);

  signals:
    /**
     * Emitted when the mailbox list is updated
     */
    void mailBoxListUpdated();

    /**
     * lets you know there are some new headers available.
     * @param mb which should be selected to see those headers
     * @param list the uid, mailbox and headers
     */
    void addHeaders( const QString& mb, const QStringList& list);

    /**
     * The body is fetched and available for retrieval
     */
    void bodyAvailable(const QString& mb, int uid);

    /**
     * This is emitted when all uids of the @p mb mailbox are known
     * you should review if the view is accurate
     */
    void allUidsKnown(const QString& mb);

    /**
     * We know the amount of messages(@p amount) in the @p mb
     */
    void messageCount(const QString& mb, const QString& amount);

    /**
     * We know the amount of unseen messages(@p amount) in the @p mb
     */
    void unseenCount(const QString& mb, int amount);

    /**
     * Reports about the network status
     */
    void status(const QString&);

    /**
     * Reports about the network status done
     */
    void statusReady();

    /**
     * Reports about the network status, there has been an error, but
     * this class is done with it, so restore the cursor and do something
     */
    void statusError(const QString&);

    /**
     * Reports about the network status, there has been an error. You need
     * to destroy the class after this error.
     */
    void error(const QString&);

    /**
     * Reports that we have an active connection
     */
    void connected();

    /**
     * Response to the previous capable() request.
     */
    void capable(const QString&, bool);

  private slots:
    void slotRead(const QString&);
    void slotError(const QString& error);
    void slotProcessQueue();
    void slotLogin();
    void slotTLS();
    void slotSendNoop();
    void slotShowDisConnectedError();

    void slotParseGetMailBoxList();
    void slotParseGetRecent();
    void slotParseMove();
    void slotParseExpunge();
    void slotParseCapabilities();
    void slotParseCheckMail();
    void slotParseExists();
    void slotParseGetHeaderList();
    void slotParseGetMessage();
    void slotParseSaveMessage();
    void slotParseCreateMailBox();
    void slotParseDeleteMailBox();
    void slotParseRenameMailBox();
};

}

#endif // Imap_H

